@@ -3,13 +3,14 @@
 #include <string.h>
 #include <math.h>
 
-#define FUNC_gcd_ui
 #include "ptypes.h"
 #include "factor.h"
-#include "util.h"
 #include "sieve.h"
 #include "mulmod.h"
 #include "cache.h"
+#include "primality.h"
+#define FUNC_gcd_ui
+#include "util.h"
 
 /*
  * You need to remember to use UV for unsigned and IV for signed types that
@@ -160,7 +161,7 @@ int trial_factor(UV n, UV *factors, UV maxtrial)
   if (maxtrial == 0)  maxtrial = UV_MAX;
 
   /* Cover the cases 0/1/2/3 now */
-  if ( (n < 4) || (maxtrial < 4) ) {
+  if (n < 4 || maxtrial < 2) {
     factors[0] = n;
     return 1;
   }
@@ -184,15 +185,17 @@ int trial_factor(UV n, UV *factors, UV maxtrial)
   /* Use the table of small primes to quickly do trial division. */
   {
     UV sp = 5;
+    UV slimit = (limit < 2003) ? limit : 2003;
     f = primes_small[sp];
-    while (f <= limit && f <= 2003) {
+    while (f <= slimit) {
       if ( (n%f) == 0 ) {
         do {
           factors[nfactors++] = f;
           n /= f;
         } while ( (n%f) == 0 );
         newlimit = isqrt(n);
-        if (newlimit < limit)  limit = newlimit;
+        if (newlimit < slimit)  slimit = newlimit;
+        if (newlimit < limit)   limit = newlimit;
       }
       f = primes_small[++sp];
     }
@@ -289,414 +292,56 @@ static int is_perfect_square(UV n, UV* sqrtn)
   return 1;
 }
 
-static int jacobi_iu(IV in, UV m) {
-  int j = 1;
-  UV n = (in < 0) ? -in : in;
 
-  if (m <= 0 || (m%2) == 0) return 0;
-  if (in < 0 && (m%4) == 3) j = -j;
-  while (n != 0) {
-    while ((n % 2) == 0) {
-      n >>= 1;
-      if ( (m % 8) == 3 || (m % 8) == 5 )  j = -j;
-    }
-    { UV t = n; n = m; m = t; }
-    if ( (n % 4) == 3 && (m % 4) == 3 )  j = -j;
-    n = n % m;
-  }
-  return (m == 1) ? j : 0;
-}
-
-
-/* Fermat pseudoprime */
-int _XS_is_pseudoprime(UV n, UV a)
-{
-  UV x;
-  UV const nm1 = n-1;
-
-  if (n == 2 || n == 3)  return 1;
-  if (n < 5) return 0;
-  if (a < 2) croak("Base %"UVuf" is invalid", a);
-  if (a >= n) {
-    a %= n;
-    if ( a <= 1 || a == nm1 )
-      return 1;
-  }
-
-  x = powmod(a, nm1, n);    /* x = a^(n-1) mod n */
-  return (x == 1);
-}
-
-
-/* Miller-Rabin probabilistic primality test
- * Returns 1 if probably prime relative to the bases, 0 if composite.
- * Bases must be between 2 and n-2
+/*
+ * The usual method, on OEIS for instance, is:
+ *    (p^(k*(e+1))-1) / (p^k-1)
+ * but that overflows quicky.  Instead we rearrange as:
+ *    1 + p^k + p^k^2 + ... p^k^e
  */
-int _XS_miller_rabin(UV n, const UV *bases, int nbases)
-{
-  UV const nm1 = n-1;
-  UV d = n-1;
-  int b, r, s = 0;
-
-  MPUassert(n > 3, "MR called with n <= 3");
-
-  while ( (d&1) == 0 ) {
-    s++;
-    d >>= 1;
-  }
-  for (b = 0; b < nbases; b++) {
-    UV x, a = bases[b];
-
-    if (a < 2)
-      croak("Base %"UVuf" is invalid", a);
-    if (a >= n)
-      a %= n;
-    if ( (a <= 1) || (a == nm1) )
-      continue;
-
-    /* n is a strong pseudoprime to this base if either
-     *   -  a^d = 1 mod n
-     *   -  a^(d2^r) = -1 mod n for some r: 0 <= r <= s-1
-     */
-
-    x = powmod(a, d, n);
-    if ( (x == 1) || (x == nm1) )  continue;
-
-    /* cover r = 1 to s-1, r=0 was just done */
-    for (r = 1; r < s; r++) {
-      x = sqrmod(x, n);
-      if ( x == nm1 )  break;
-      if ( x == 1   )  return 0;
-    }
-    if (r >= s)
-      return 0;
-  }
-  return 1;
-}
-
-/* M-R with a = 2 and some checks removed.  For internal use. */
-int _SPRP2(UV n)
-{
-  UV const nm1 = n-1;
-  UV d = n-1;
-  UV x;
-  int r, s = 0;
-
-  MPUassert(n > 3, "S-PRP-2 called with n <= 3");
-  if (!(n & 1)) return 0;
-  while ( (d & 1) == 0 ) {  s++;  d >>= 1; }
-  /* n is a strong pseudoprime to this base if either
-   *   -  a^d = 1 mod n
-   *   -  a^(d2^r) = -1 mod n for some r: 0 <= r <= s-1 */
-  x = powmod(2, d, n);
-  if (x == 1 || x == nm1)  return 1;
-
-  /* just did r=0, now test r = 1 to s-1 */
-  for (r = 1; r < s; r++) {
-    x = sqrmod(x, n);
-    if (x == nm1)  return 1;
-  }
-  return 0;
-}
-
-/* Select M-R bases from http://miller-rabin.appspot.com/, 27 May 2013 */
-#if BITS_PER_WORD == 32
-static const UV mr_bases_small_2[2] = {31, 73};
-static const UV mr_bases_small_3[3] = {2, 7, 61};
-#else
-static const UV mr_bases_large_1[1] = { UVCONST(  9345883071009581737 ) };
-static const UV mr_bases_large_2[2] = { UVCONST(   725270293939359937 ),
-                                        UVCONST(  3569819667048198375 ) };
-static const UV mr_bases_large_3[3] = { UVCONST(  4230279247111683200 ),
-                                        UVCONST( 14694767155120705706 ),
-                                        UVCONST( 16641139526367750375 ) };
-#endif
-
-int _XS_is_prob_prime(UV n)
-{
-  int ret;
-
-  if (n < 11) {
-    if (n == 2 || n == 3 || n == 5 || n == 7)     return 2;
-    else                                          return 0;
-  }
-  if (!(n%2) || !(n%3) || !(n%5) || !(n%7))       return 0;
-  if (n <  121) /* 11*11 */                       return 2;
-  if (!(n%11) || !(n%13) || !(n%17) || !(n%19) ||
-      !(n%23) || !(n%29) || !(n%31) || !(n%37) ||
-      !(n%41) || !(n%43) || !(n%47) || !(n%53))   return 0;
-  if (n < 3481) /* 59*59 */                       return 2;
-
-#if BITS_PER_WORD == 32
-  /* We could use one base when n < 49191, two when n < 360018361. */
-  if (n < UVCONST(9080191))
-    ret = _XS_miller_rabin(n, mr_bases_small_2, 2);
-  else
-    ret = _XS_miller_rabin(n, mr_bases_small_3, 3);
-#else
-  /* AESLSP test costs about 1.5 Selfridges, vs. ~2.2 for strong Lucas.
-   * So it works out to be faster to do AES-BPSW vs. 3 M-R tests. */
-  if (n < UVCONST(341531))
-    ret = _XS_miller_rabin(n, mr_bases_large_1, 1);
-  else if (n < UVCONST(885594169))
-    ret = _XS_miller_rabin(n, mr_bases_large_2, 2);
-  else
-    ret = _SPRP2(n) && _XS_is_almost_extra_strong_lucas_pseudoprime(n,1);
-#endif
-  return 2*ret;
-}
-
-/* Generic Lucas sequence for any appropriate P and Q */
-void lucas_seq(UV* Uret, UV* Vret, UV* Qkret, UV n, IV P, IV Q, UV k)
-{
-  UV U, V, b, Dmod, Qmod, Pmod, Qk;
-
-  if (k == 0) {
-    *Uret = 0;
-    *Vret = 2;
-    *Qkret = Q;
-    return;
-  }
-
-  Qmod = (Q < 0)  ?  (UV)(Q + (IV)n)  :  (UV)Q;
-  Pmod = (P < 0)  ?  (UV)(P + (IV)n)  :  (UV)P;
-  Dmod = submod( mulmod(Pmod, Pmod, n), mulmod(4, Qmod, n), n);
-  MPUassert(Dmod != 0, "lucas_seq: D is 0");
-  U = 1;
-  V = Pmod;
-  Qk = Qmod;
-  { UV v = k; b = 1; while (v >>= 1) b++; }
-
-  if (Q == 1) {
-    while (b > 1) {
-      U = mulmod(U, V, n);
-      V = mulsubmod(V, V, 2, n);
-      b--;
-      if ( (k >> (b-1)) & UVCONST(1) ) {
-        UV t2 = mulmod(U, Dmod, n);
-        U = muladdmod(U, Pmod, V, n);
-        if (U & 1) { U = (n>>1) + (U>>1) + 1; } else { U >>= 1; }
-        V = muladdmod(V, P, t2, n);
-        if (V & 1) { V = (n>>1) + (V>>1) + 1; } else { V >>= 1; }
-      }
-    }
-  } else if (P == 1 && Q == -1) {
-    /* This is about 30% faster than the generic code below.  Since 50% of
-     * Lucas and strong Lucas tests come here, I think it's worth doing. */
-    int sign = Q;
-    while (b > 1) {
-      U = mulmod(U, V, n);
-      if (sign == 1) V = mulsubmod(V, V, 2, n);
-      else           V = muladdmod(V, V, 2, n);
-      sign = 1;   /* Qk *= Qk */
-      b--;
-      if ( (k >> (b-1)) & UVCONST(1) ) {
-        UV t2 = mulmod(U, Dmod, n);
-        U = addmod(U, V, n);
-        if (U & 1) { U = (n>>1) + (U>>1) + 1; } else { U >>= 1; }
-        V = addmod(V, t2, n);
-        if (V & 1) { V = (n>>1) + (V>>1) + 1; } else { V >>= 1; }
-        sign = -1;  /* Qk *= Q */
-      }
-    }
-    if (sign == 1) Qk = 1;
-  } else {
-    while (b > 1) {
-      U = mulmod(U, V, n);
-      V = mulsubmod(V, V, addmod(Qk,Qk,n), n);
-      Qk = sqrmod(Qk, n);
-      b--;
-      if ( (k >> (b-1)) & UVCONST(1) ) {
-        UV t2 = mulmod(U, Dmod, n);
-        U = muladdmod(U, Pmod, V, n);
-        if (U & 1) { U = (n>>1) + (U>>1) + 1; } else { U >>= 1; }
-        V = muladdmod(V, P, t2, n);
-        if (V & 1) { V = (n>>1) + (V>>1) + 1; } else { V >>= 1; }
-        Qk = mulmod(Qk, Qmod, n);
-      }
-    }
-  }
-  *Uret = U;
-  *Vret = V;
-  *Qkret = Qk;
-}
-
-/* Lucas tests:
- *  0: Standard
- *  1: Strong
- *  2: Extra Strong (Mo/Jones/Grantham)
- *
- * Goal:
- *       (1) no false results when combined with the SPRP-2 test.
- *       (2) fast enough to use SPRP-2 + this in place of 3+ M-R tests.
- *
- * For internal purposes, we typically want to use the extra strong test
- * because it is slightly faster (Q = 1 simplies things).  None of them have
- * any false positives for the BPSW test.
- *
- * This runs 4-7x faster than MPU::GMP, which means ~10x faster than most GMP
- * implementations.  It is about 2x slower than a single M-R test.
- */
-int _XS_is_lucas_pseudoprime(UV n, int strength)
-{
-  IV P, Q, D;
-  UV U, V, Qk, d, s;
-
-  if (n == 2 || n == 3) return 1;
-  if (n < 5 || (n%2) == 0) return 0;
-  if (n == UV_MAX) return 0;
-
-  if (strength < 2) {
-    UV Du = 5;
-    IV sign = 1;
-    while (1) {
-      D = Du * sign;
-      if (gcd_ui(Du, n) > 1 && gcd_ui(Du, n) != n) return 0;
-      if (jacobi_iu(D, n) == -1)
-        break;
-      if (Du == 21 && is_perfect_square(n, 0)) return 0;
-      Du += 2;
-      sign = -sign;
-    }
-    P = 1;
-    Q = (1 - D) / 4;
-  } else {
-    P = 3;
-    Q = 1;
-    while (1) {
-      D = P*P - 4;
-      if (gcd_ui(D, n) > 1 && gcd_ui(D, n) != n) return 0;
-      if (jacobi_iu(D, n) == -1)
-        break;
-      if (P == 21 && is_perfect_square(n, 0)) return 0;
-      P++;
-    }
-  }
-  MPUassert( D == (P*P - 4*Q) , "is_lucas_pseudoprime: incorrect DPQ");
-
-  d = n+1;
-  s = 0;
-  if (strength > 0)
-    while ( (d & 1) == 0 ) {  s++;  d >>= 1; }
-  lucas_seq(&U, &V, &Qk, n, P, Q, d);
-
-  if (strength == 0) {
-    if (U == 0)
-      return 1;
-  } else if (strength == 1) {
-    if (U == 0)
-      return 1;
-    /* Now check to see if V_{d*2^r} == 0 for any 0 <= r < s */
-    while (s--) {
-      if (V == 0)
-        return 1;
-      if (s) {
-        V = mulsubmod(V, V, addmod(Qk,Qk,n), n);
-        Qk = sqrmod(Qk, n);
-      }
-    }
-  } else {
-    if ( U == 0 && (V == 2 || V == (n-2)) )
-      return 1;
-    /* Now check to see if V_{d*2^r} == 0 for any 0 <= r < s-1 */
-    s--;
-    while (s--) {
-      if (V == 0)
-        return 1;
-      if (s)
-        V = mulsubmod(V, V, 2, n);
-    }
-  }
-  return 0;
-}
-
-/* A generalization of Pari's shortcut to the extra-strong Lucas test.
- * I've added a gcd check at the top, which needs to be done and also results
- * in fewer pseudoprimes.  Pari always does trial division to 100 first so
- * is unlikely to come up there.  This only calculate V, which can be done
- * faster, but that means we have more pseudoprimes than the standard
- * extra-strong test.
- *
- * increment:  1 for Baillie OEIS, 2 for Pari.
- *
- * With increment = 1, these results will be a subset of the extra-strong
- * Lucas pseudoprimes.  With increment = 2, we produce Pari's results.
- */
-int _XS_is_almost_extra_strong_lucas_pseudoprime(UV n, UV increment)
-{
-  UV P, V, d, s;
-
-  if (n == 2 || n == 3 || n == 5) return 1;
-  if (n < 7 || (n%2) == 0) return 0;
-  if (n == UV_MAX) return 0;
-  if (increment < 1 || increment > 256)
-    croak("Invalid lucas paramater increment: %"UVuf"\n", increment);
-
-  P = 3;
-  while (1) {
-    UV D = P*P - 4;
-    d = gcd_ui(D, n);
-    if (d > 1 && d < n)
-      return 0;
-    if (jacobi_iu(D, n) == -1)
-      break;
-    if (P == (3+20*increment) && is_perfect_square(n, 0)) return 0;
-    P += increment;
-    if (P > 65535)
-      croak("lucas_extrastrong_params: P exceeded 65535");
-  }
-  if (P >= n)  P %= n;   /* Never happens with increment < 4 */
-
-  d = n+1;
-  s = 0;
-  while ( (d & 1) == 0 ) {  s++;  d >>= 1; }
-
-  {
-    UV W, b;
-    V = P;
-    W = mulsubmod(P, P, 2, n);
-    { UV v = d; b = 1; while (v >>= 1) b++; }
-    while (b-- > 1) {
-      if ( (d >> (b-1)) & UVCONST(1) ) {
-        V = mulsubmod(V, W, P, n);
-        W = mulsubmod(W, W, 2, n);
-      } else {
-        W = mulsubmod(V, W, P, n);
-        V = mulsubmod(V, V, 2, n);
-      }
-    }
-  }
-
-  if (V == 2 || V == (n-2))
-    return 1;
-  while (s-- > 1) {
-    if (V == 0)
-      return 1;
-    V = mulsubmod(V, V, 2, n);
-    if (V == 2)
-      return 0;
-  }
-  return 0;
-}
-
-
-UV _XS_divisor_sum(UV n)
+UV _XS_divisor_sum(UV n, UV k)
 {
   UV factors[MPU_MAX_FACTORS+1];
-  int nfac, i;
+  int nfac, i, j;
   UV product = 1;
 
   if (n <= 1) return n;
   nfac = factor(n, factors);
-  for (i = 0; i < nfac; i++) {
-    if (i+1 < nfac && factors[i] == factors[i+1]) {
-      UV fmult = factors[i]*factors[i];
-      do {
-        fmult *= factors[i++];
-      } while (i+1 < nfac && factors[i] == factors[i+1]);
-      product *= (fmult-1) / (factors[i]-1);
-    } else {
-      product *= factors[i]+1;
+  if (k == 0) {
+    for (i = 0; i < nfac; i++) {
+      UV e = 1,  f = factors[i];
+      while (i+1 < nfac && f == factors[i+1]) { e++; i++; }
+      product *= (e+1);
+    }
+  } else if (k == 1) {
+    for (i = 0; i < nfac; i++) {
+      UV e = 1,  f = factors[i];
+      UV fmult = 1 + f;
+      while (i+1 < nfac && f == factors[i+1]) { e++; i++; }
+      if (e > 1) {
+        UV pke = f;
+        for (j = 1; j < (int)e; j++) {
+          pke *= f;
+          fmult += pke;
+        }
+      }
+      product *= fmult;
+    }
+  } else {
+    for (i = 0; i < nfac; i++) {
+      UV e = 1,  f = factors[i];
+      UV fmult, pk = f;
+      for (j = 1; j < (int)k; j++)  pk *= f;
+      while (i+1 < nfac && f == factors[i+1]) { e++; i++; }
+      fmult = 1 + pk;
+      if (e > 1) {
+        UV pke = pk;
+        for (j = 1; j < (int)e; j++) {
+          pke *= pk;
+          fmult += pke;
+        }
+      }
+      product *= fmult;
     }
   }
   return product;
@@ -1378,135 +1023,3 @@ int racing_squfof_factor(UV n, UV *factors, UV rounds)
   factors[0] = n;
   return 1;
 }
-
-
-/****************************************************************************/
-
-/*
- *
- * The Frobenius-Underwood test has no known counterexamples below 10^13, but
- * has not been extensively tested above that.  This is the Minimal Lambda+2
- * test from section 9 of "Quadratic Composite Tests" by Paul Underwood.
- *
- * Given the script:
- *  time mpu 'forprimes { Math::Prime::Util::_XS_is_frobenius_underwood_pseudoprime($_); Math::Prime::Util::_XS_is_frobenius_underwood_pseudoprime($_+2); } 500_000_000'
- * and replacing the tests appropriately, I get these times:
- *
- *    0.87    $_ (cost of empty loop)
- *   21.37    _XS_is_pseudoprime($_,2)
- *   22.42    _XS_miller_rabin($_,2)
- *   44.53    _XS_is_lucas_pseudoprime($_)
- *   43.95    _XS_is_strong_lucas_pseudoprime($_)
- *   40.09    _XS_is_extra_strong_lucas_pseudoprime($_)
- *   25.86    _XS_is_almost_extra_strong_lucas_pseudoprime($_)
- *   42.40    _XS_is_frobenius_underwood_pseudoprime($_)
- *   27.02    _XS_is_prob_prime($_)
- *   27.24    _XS_is_prime($_)
- *
- * At these sizes is_prob_prime is doing 1-2 M-R tests.  The input validation
- * is adding a noticeable overhead to is_prime.
- *
- * With a set of 100k 64-bit random primes; 'do { die unless ... } for 1..50'
- *
- *   0.32    empty loop
- *  10.25    _XS_is_pseudoprime($_,2)
- *  10.06    _XS_miller_rabin($_,2)
- *  22.02    _XS_is_lucas_pseudoprime($_)
- *  21.81    _XS_is_strong_lucas_pseudoprime($_)
- *  20.99    _XS_is_extra_strong_lucas_pseudoprime($_)
- *  14.01    _XS_is_almost_extra_strong_lucas_pseudoprime($_)
- *  18.44    _XS_is_frobenius_underwood_pseudoprime($_)
- *  24.11    _XS_is_prob_prime($_)
- *  24.06    _XS_is_prime($_)
- *
- * At this point is_prob_prime has transitioned to BPSW.
- *
- * Calling a powmod a 'Selfridge' unit, then we see:
- *    1   Selfridge unit    M-R test
- *    1.4 Selfridge unit    "almost extra strong" Lucas
- *    2   Selfridge units   Lucas or Frobenius-Underwood
- *    3   Selfridge units   BPSW (standard, strong, or extra-strong)
- *
- * We try to structure the primality test like:
- *   1) simple divisibility    very fast       primes and ~10% of composites
- *   2) M-R with base 2        1 Selfridge     primes and .00000000002% comps
- *   3) Lucas test             2 Selfridge     only primes
- *
- * Hence given a random composite, about 90% of the time it costs us almost
- * nothing.  After spending 1 Selfridge on the first MR test, less than 32M
- * composites remain undecided out of 18 quintillion 64-bit composites.  The
- * final Lucas test has no false positives.
- * Replacing the Lucas test with the F-U test won't save any time.  Replacing
- * the whole thing with the F-U test (assuming it has no false results for
- * all 64-bit values, which hasn't been verified), doesn't help either.
- * It's 2/3 the cost for primes, but much more expensive for composites.  It
- * seems of interest for > 2^64 as a different test to do in addition to BPSW.
- */
-
-
-int _XS_is_frobenius_underwood_pseudoprime(UV n)
-{
-  int bit;
-  UV x, result, multiplier, a, b, np1, len, t1, t2, na;
-  IV t;
-
-  if (n < 2) return 0;
-  if (n < 4) return 1;
-  if ((n % 2) == 0) return 0;
-  if (is_perfect_square(n,0)) return 0;
-  if (n == UV_MAX) return 0;
-
-  x = 0;
-  t = -1;
-  while ( jacobi_iu( t, n ) != -1 ) {
-    x++;
-    t = (IV)(x*x) - 4;
-  }
-  result = addmod( addmod(x, x, n), 5, n);
-  multiplier = addmod(x, 2, n);
-
-  a = 1;
-  b = 2;
-  np1 = n+1;
-  { UV v = np1; len = 1;  while (v >>= 1) len++; }
-
-  if (x == 0) {
-    for (bit = len-2; bit >= 0; bit--) {
-      t2 = addmod(b, b, n);
-      na = mulmod(a, t2, n);
-      t1 = addmod(b, a, n);
-      t2 = submod(b, a, n);
-      b = mulmod(t1, t2, n);
-      a = na;
-      if ( (np1 >> bit) & UVCONST(1) ) {
-        t1 = mulmod(a, 2, n);
-        na = addmod(t1, b, n);
-        t1 = addmod(b, b, n);
-        b = submod(t1, a, n);
-        a = na;
-      }
-    }
-  } else {
-    for (bit = len-2; bit >= 0; bit--) {
-      t1 = mulmod(a, x, n);
-      t2 = addmod(b, b, n);
-      t1 = addmod(t1, t2, n);
-      na = mulmod(a, t1, n);
-      t1 = addmod(b, a, n);
-      t2 = submod(b, a, n);
-      b = mulmod(t1, t2, n);
-      a = na;
-      if ( (np1 >> bit) & UVCONST(1) ) {
-        t1 = mulmod(a, multiplier, n);
-        na = addmod(t1, b, n);
-        t1 = addmod(b, b, n);
-        b = submod(t1, a, n);
-        a = na;
-      }
-    }
-  }
-  if (_XS_get_verbose()>1) printf("%"UVuf" is %s with x = %"UVuf"\n", n, (a == 0 && b == result) ? "probably prime" : "composite", x);
-  if (a == 0 && b == result)
-    return 1;
-  return 0;
-}
