@@ -4,7 +4,7 @@ use warnings;
 
 BEGIN {
   $Math::Prime::Util::PrimeIterator::AUTHORITY = 'cpan:DANAJ';
-  $Math::Prime::Util::PrimeIterator::VERSION = '0.44_004';
+  $Math::Prime::Util::PrimeIterator::VERSION = '0.32';
 }
 
 use base qw( Exporter );
@@ -13,63 +13,59 @@ our %EXPORT_TAGS = (all => [ @EXPORT_OK ]);
 
 
 use Math::Prime::Util qw/next_prime prev_prime is_prime prime_count nth_prime/;
+use Math::BigInt try => "GMP,Pari";
 
-# We're going to use a scalar rather than a hash because there is currently
-# only one data object (the current value) and this makes it little faster.
+my $bigprime = Math::BigInt->new(
+  (~0 == 4294967295) ? "4294967311" : "18446744073709551629"
+);
 
 sub new {
   my ($class, $start) = @_;
-  my $p = 2;
-  my $self = bless \$p, $class;
+  my $self = bless {
+    p => 2,
+  }, $class;
   $self->rewind($start) if defined $start;
   return $self;
 }
 
-# To make Iterator::Simple happy.
-sub __iter__ {
-  my $self = shift;
-  require Iterator::Simple;
-  return Iterator::Simple::iterator(sub { $self->iterate });
-  $self;
-}
-
-sub value { ${$_[0]}; }
+sub value { $_[0]->{p}; }
 sub next {
-  #my $self = shift;  $$self = next_prime($$self);  return $self;
-  ${$_[0]} = next_prime(${$_[0]});
-  return $_[0];
+  my $self = shift;
+  my $np = next_prime($self->{p}) || $bigprime;
+  $self->{p} = $np;
+  return $self;
 }
 sub prev {
   my $self = shift;
-  my $p = $$self;
-  $$self = ($p <= 2) ? 2 : prev_prime($p);
+  my $p = $self->{p};
+  $self->{p} = ($p <= 2) ? 2 : prev_prime($p);
   return $self;
 }
 sub iterate {
-  #my $self = shift;  my $p = $$self;  $$self = next_prime($p);  return $p;
-  my $p = ${$_[0]};
-  ${$_[0]} = next_prime(${$_[0]});
+  my $self = shift;
+  my $p = $self->{p};
+  my $np = next_prime($p) || $bigprime;
+  $self->{p} = $np;
   return $p;
 }
 
 sub rewind {
   my ($self, $start) = @_;
-  $$self = 2;
+  $self->{p} = 2;
   if (defined $start && $start ne '2') {
     Math::Prime::Util::_validate_num($start)
       || Math::Prime::Util::_validate_positive_integer($start);
-    $$self = next_prime($start-1) if $start > 2;
+    if ($start > 2) {
+      $self->{p} = next_prime($start-1) || $bigprime;
+    }
   }
   return $self;
 }
 
-sub peek {
-  return next_prime(${$_[0]});
-}
-
 # Some methods to match Math::NumSeq
 sub tell_i {
-  return prime_count(${$_[0]});
+  my $self = shift;
+  return prime_count($self->{p});
 }
 sub pred {
   my($self, $n) = @_;
@@ -120,8 +116,6 @@ __END__
 
 =for stopwords prev pred ith i'th
 
-=for test_synopsis use v5.14;  my ($i,$n) = (2,2);
-
 =head1 NAME
 
 Math::Prime::Util::PrimeIterator - An object iterator for primes
@@ -129,7 +123,7 @@ Math::Prime::Util::PrimeIterator - An object iterator for primes
 
 =head1 VERSION
 
-Version 0.44_004
+Version 0.32
 
 
 =head1 SYNOPSIS
@@ -138,13 +132,13 @@ Version 0.44_004
   my $it = Math::Prime::Util::PrimeIterator->new();
 
   # Simple use: return current value and move forward.
-  my $sum = 0;  $sum += $it->iterate() for 1..10000;
+  $sum += $it->iterate() for 1..10000;
 
   # Methods
   my $v = $it->value();     # Return current value
   $it->next();              # Move to next prime (returns self)
   $it->prev();              # Move to prev prime (returns self)
-  $v = $it->iterate();      # Returns current value; moves to next prime
+  my $v = $it->iterate();   # Returns current value and moves to next prime
   $it->rewind();            # Resets position to 2
   $it->rewind($n);          # Resets position to next_prime($n-1)
 
@@ -152,8 +146,8 @@ Version 0.44_004
   $it->tell_i();            # Returns the index of the current position
   $it->pred($n);            # Returns true if $n is prime
   $it->ith($i);             # Returns the $ith prime
-  $it->value_to_i($n);      # Returns the index of the first prime >= $n
-  $it->value_to_i_estimate($n);  # Approx index of value $n
+  $it->value_to_i($n)       # Returns the index of the first prime >= $n
+  $it->value_to_i_estimate($n)  # Approx index of value $n
 
   # Methods similar to Math::NumSeq, changes iterator
   $it->seek_to_i($i);       # Resets position to the $ith prime
@@ -200,10 +194,6 @@ the next prime.
 Resets the current position to either 2 or, if given an integer argument,
 the least prime not less than the argument.
 
-=head2 peek
-
-Returns the value at the next position without moving the iterator.
-
 =head2 tell_i
 
 Returns the index of the current position, starting at 1 (corresponding to
@@ -274,8 +264,6 @@ L<Math::Prime::Util::PrimeArray>
 
 L<Math::NumSeq::Primes>
 
-L<List::Gen>
-
 =head1 AUTHORS
 
 Dana Jacobsen E<lt>dana@acm.orgE<gt>
